
import unittest
from math import exp, sqrt

import numpy as np

from finsim.simulation.stock import BlackScholesMertonStockPrices, HestonStockPrices, MertonJumpDiffusionStockPrices


def expected_bsm_stock(S0, r, sigma, t):
    expected_S = S0 * exp(r*t)
    variance_S = S0 * S0 * (exp(sigma*sigma*t)-1) * exp((2*r+sigma*sigma)*t)
    return expected_S, variance_S


class TestStockSimulations(unittest.TestCase):
    def backend_test_BlackScholesMertonStocks(self, S, r, sigma, T, dt, nbsimulations):
        bsm_simulator = BlackScholesMertonStockPrices(S, r, sigma)
        sarray = bsm_simulator.generate_time_series(T, dt, nbsimulations=nbsimulations)
        average_s = np.mean(sarray[:, -1])

        expected_s, variance_s = expected_bsm_stock(S, r, sigma, T)

        self.assertAlmostEqual(average_s, expected_s, delta=1.96*sqrt(variance_s))

    def test_BlackScholesMertonStocks(self):
        self.backend_test_BlackScholesMertonStocks(100, 0.04, 0.02, 100, 0.1, 10000)

        for S in np.random.lognormal(100, 10, size=3):
            for r in np.random.lognormal(0.04, 0.01, size=3):
                for sigma in np.random.lognormal(0.01, 0.0001, size=3):
                    self.backend_test_BlackScholesMertonStocks(S, r, sigma, 100, 0.1, 10000)

    def test_HestonStocks(self):
        h_simulator = HestonStockPrices(100, 0.01, 0.01, 0.01, 0.1, 0.0001, 0.2)
        h_stocks = [
            np.mean(h_simulator.generate_time_series(100, 0.1, nbsimulations=1000)[0], axis=0)
            for _ in range(100)
        ]

        self.assertAlmostEqual(
            np.mean([stock[100] for stock in h_stocks]),
            110.52,
            delta=1.96*0.11
        )
        self.assertAlmostEqual(
            np.mean([stock[250] for stock in h_stocks]),
            128.41,
            delta=1.96*0.23
        )
        self.assertAlmostEqual(
            np.mean([stock[500] for stock in h_stocks]),
            164.92,
            delta=1.96*0.38
        )
        self.assertAlmostEqual(
            np.mean([stock[750] for stock in h_stocks]),
            211.80,
            delta=1.96*0.56
        )
        self.assertAlmostEqual(
            np.mean([stock[999] for stock in h_stocks]),
            271.72,
            delta=1.96*0.80
        )

    def test_MertonJumpDiffusionStocks(self):
        mjd_simulator = MertonJumpDiffusionStockPrices(100, 0.01, 0.01, 0.01, 0.01, 0.001)
        mjd_stocks = [
            np.mean(mjd_simulator.generate_time_series(100, 0.1, nbsimulations=1000), axis=0)
            for _ in range(100)
        ]

        self.assertAlmostEqual(
            np.mean([stock[250] for stock in mjd_stocks]),
            197.97,
            delta=1.96*6.37
        )
        self.assertAlmostEqual(
            np.mean([stock[500] for stock in mjd_stocks]),
            393.35,
            delta=1.96*28.15
        )
        self.assertAlmostEqual(
            np.mean([stock[750] for stock in mjd_stocks]),
            773.20,
            delta=1.96*67.73
        )
        self.assertAlmostEqual(
            np.mean([stock[999] for stock in mjd_stocks]),
            1523.06,
            delta=1.96*169.68
        )


if __name__ == '__main__':
    unittest.main()




from datetime import datetime
import os
import logging

import pandas as pd
from pandas_datareader import data
from pandas_datareader._utils import RemoteDataError
import tables as tb


# native yahoo reader
def extract_online_yahoofinance_data(symbol, startdate, enddate):
    try:
        df = data.DataReader(
            name=symbol,
            data_source='yahoo',
            start=datetime.strptime(startdate, '%Y-%m-%d'),
            end=datetime.strptime(enddate, '%Y-%m-%d')
        )
    except (KeyError, RemoteDataError) as e:
        logging.warning('Symbol {} does not exist between {} and {}.'.format(symbol, startdate, enddate))
        return pd.DataFrame({
            'TimeStamp': [],
            'High': [],
            'Low': [],
            'Open': [],
            'Close': [],
            'Volume': [],
            'Adj Close': []
        })

    oricols = df.columns
    df['TimeStamp'] = df.index
    # df['Date'] = df['TimeStamp'].apply(lambda ts: ts.date())
    df = df[['TimeStamp'] + list(oricols)]

    return df


# yahoo reader with local cache
METATABLE_FILENAME = 'SYMBOL_CACHE.h5'
METATABLE_ROWDES = {
    'symbol': tb.StringCol(shape=(), itemsize=10, dflt='', pos=0),
    'query_startdate': tb.StringCol(shape=(), itemsize=10, dflt='0000-00-00', pos=1),
    'query_enddate': tb.StringCol(shape=(), itemsize=10, dflt='0000-00-00', pos=2),
    'data_startdate': tb.StringCol(shape=(), itemsize=10, dflt='0000-00-00', pos=3),
    'data_enddate': tb.StringCol(shape=(), itemsize=10, dflt='0000-00-00', pos=4)
}


def get_yahoofinance_data(symbol, startdate, enddate, cacheddir=None):
    if cacheddir is None:
        return extract_online_yahoofinance_data(symbol, startdate, enddate)

    if isinstance(cacheddir, str):
        if not os.path.exists(cacheddir) or (os.path.exists(cacheddir) and not os.path.isdir(cacheddir)):
            logging.info('Creating directory: {}'.format(cacheddir))
            os.makedirs(cacheddir)
        if not os.path.exists(os.path.join(cacheddir, METATABLE_FILENAME)):
            logging.info('Creating file: {}'.format(os.path.join(cacheddir, METATABLE_FILENAME)))
            metatable_h5file = tb.open_file(os.path.join(cacheddir, METATABLE_FILENAME), 'w')
            table = metatable_h5file.create_table('/', 'metatable', METATABLE_ROWDES, title='metatable')
        else:
            metatable_h5file = tb.open_file(os.path.join(cacheddir, METATABLE_FILENAME), 'r+')
            table = metatable_h5file.root.metatable

        preexist = False
        for row in table.where('symbol=="{}"'.format(symbol)):
            preexist = True
            if row['query_startdate'].decode('utf-8') <= startdate and row['query_enddate'].decode('utf-8') >= enddate:
                df = pd.read_hdf(os.path.join(cacheddir, '{}.h5'.format(symbol)), 'yahoodata')
                if len(df) > 0:
                    df = df[(df['TimeStamp'] >= startdate) & (df['TimeStamp'] <= enddate)]
                metatable_h5file.close()
                return df

        df = extract_online_yahoofinance_data(symbol, startdate, enddate)
        logging.debug('Caching data for {} from {} to {}'.format(symbol, startdate, enddate))
        df.to_hdf(os.path.join(cacheddir, '{}.h5'.format(symbol)), 'yahoodata')

        if preexist:
            logging.debug('Updating symbol {} in metatable.'.format(symbol))
            for row in table.where('symbol=="{}"'.format(symbol)):
                row['query_startdate'] = startdate
                row['query_enddate'] = enddate
                if len(df) > 0:
                    row['data_startdate'] = datetime.strftime(df['TimeStamp'][0].date(), '%Y-%m-%d')
                    row['data_enddate'] = datetime.strftime(df['TimeStamp'][-1].date(), '%Y-%m-%d')
                else:
                    row['data_startdate'] = '0000-00-00'
                    row['data_enddate'] = ' 0000-00-00'
                row.update()
        else:
            logging.debug('Creating symbol {} in metatable'.format(symbol))
            newrow = table.row
            newrow['symbol'] = symbol
            newrow['query_startdate'] = startdate
            newrow['query_enddate'] = enddate
            if len(df) > 0:
                newrow['data_startdate'] = datetime.strftime(df['TimeStamp'][0].date(), '%Y-%m-%d')
                newrow['data_enddate'] = datetime.strftime(df['TimeStamp'][-1].date(), '%Y-%m-%d')
            else:
                newrow['data_startdate'] = '0000-00-00'
                newrow['data_enddate'] = '0000-00-00'
            newrow.append()

        table.flush()
        metatable_h5file.close()

        return df
    else:
        raise TypeError('Type of cacheddir has to be str, but got {} instead!'.format(type(cacheddir)))
